SystemOrganization addCategory: #'Xtreams-Core'!

!Object methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:12'!
streamingInsert: anInteger into: aWriteStream
	anInteger timesRepeat: [self streamingInsertInto: aWriteStream]! !

!Object methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:12'!
streamingInsertInto: aWriteStream
	aWriteStream insert: (aWriteStream contentsSpecies with: self)! !

!Object methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:12'!
streamingPrintOn: aWriteStream
	aWriteStream write: self printString! !

!Object methodsFor: '*Xtreams-Core-private' stamp: 'MartinKobetic 12/22/2010 16:43'!
streamingWrite: anInteger into: aWriteStream
	anInteger timesRepeat: [self streamingWriteInto: aWriteStream].
	^anInteger! !

!Object methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:12'!
streamingWriteInto: aWriteStream
	aWriteStream put: self! !

Object subclass: #XTBuffer
	instanceVariableNames: 'cache readPosition writePosition dataLength'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Core'!

!XTBuffer commentStamp: '<historical>' prior: 0!
Buffer implements a buffering API over an in-memory cache. The cache will grow as required to fit new data in to the buffer. Two positions are kept, the read position and write position, allowing flexible usage of the buffer.

API:
	read:into:startingAt: -- reads data from the cache and moves readPosition forward
	write:into:startingAt: -- writes data into the cache and moves writePosition forward
	readSkip:/writeSkip: -- moves either readPosition or writePosition forward without changing the cache
	readPosition/readPosition:/writePosition/writePosition: -- accessors to the read/write positions
	clear -- remove all the data from the cache
	trim -- remove all the read data and all the empty space for writing in to from the cache

Instance Variables
	cache	<SequenceableCollection>	the cache for our buffer
	dataLength	<ArithmeticValue>	the amount of data in our cache, not the size of the cache
	readPosition	<ArithmeticValue>	the position within our data that we're reading from (0..dataLength)
	writePosition	<ArithmeticValue>	the position within our data that we're writing from (0..dataLength)

!

!XTBuffer classSide methodsFor: 'instance creation' stamp: 'MartinKobetic 12/30/2010 21:46'!
new: anInteger class: aClass
	^self on: (aClass newRecycled: anInteger)! !

!XTBuffer classSide methodsFor: 'instance creation' stamp: 'nice 10/9/2010 18:42'!
on: aSequenceableCollection
	^self new on: aSequenceableCollection! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:29'!
activeSize
	"The number of elements in the buffer that can be explored. This is <= cacheSize"

	^dataLength! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:29'!
cache
	^cache! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:29'!
cacheSize
	"The size of the cache, which is >= dataLength"

	^cache size! !

!XTBuffer methodsFor: 'compressing' stamp: 'nice 10/9/2010 18:38'!
clear
	self clearCache.
	readPosition := writePosition := dataLength := 0! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:40'!
clearCache
	cache := cache copyEmpty: 0! !

!XTBuffer methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 18:39'!
close
	dataLength := writePosition! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:32'!
contentsFuture
	^cache copyFrom: readPosition + 1 to: readPosition + self readSize! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:32'!
contentsPast
	^cache copyFrom: 1 to: writePosition! !

!XTBuffer methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 18:39'!
contentsSpecies
	^cache species! !

!XTBuffer methodsFor: 'accessing - reading' stamp: 'MartinKobetic 1/11/2011 17:53'!
get
	dataLength = readPosition ifTrue: [Incomplete zero raise].
	readPosition := readPosition + 1.
	^cache at: readPosition! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:40'!
growBy: anInteger
	| replacement |
	replacement := cache copyEmpty: (cache size + anInteger) * 2.
	replacement replaceFrom: 1 to: cache size with: cache startingAt: 1.
	cache := replacement! !

!XTBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 18:41'!
hasDataToRead
	^readPosition < dataLength! !

!XTBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 18:41'!
hasFixedWriteSpace
	^false! !

!XTBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 18:41'!
hasSpaceToWrite
	^true! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:33'!
inactiveSize
	"Free space to write in to"

	^cache size - dataLength! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:39'!
insert: aStreamable

	^aStreamable streamingInsertInto: self! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:36'!
insert: anInteger from: aStreamable

	^aStreamable streamingInsert: anInteger into: self! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:39'!
insert: anInteger from: aSequenceableCollection at: startIndex
	"Implementation note: this code does a memory move to the right.
	Since Squeak primitive #replaceFrom:to:with: is implemented with a memcpy rather than a memmove, it cannot work.
	We have to buffer the copy in a motionBuffer temporary."

	| motionBuffer |
	motionBuffer := cache copyFrom: writePosition + 1 to: dataLength.
	dataLength + anInteger <= cache size ifFalse: [self growBy: anInteger].

	cache replaceFrom: writePosition + anInteger + 1 to: dataLength + anInteger with: motionBuffer startingAt: 1.
	cache replaceFrom: writePosition + 1 to: writePosition + anInteger with: aSequenceableCollection startingAt: startIndex.
	writePosition := writePosition + anInteger.
	dataLength := dataLength + anInteger.
	^anInteger! !

!XTBuffer methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 18:39'!
on: aSequenceableCollection
	cache := aSequenceableCollection.
	dataLength := 0.
	readPosition := 0.
	writePosition := 0! !

!XTBuffer methodsFor: 'printing' stamp: 'nice 10/9/2010 18:39'!
printOn: aStream
	| stream |
	stream := String new writing.
	self streamingPrintOn: stream.
	aStream nextPutAll: stream conclusion! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:40'!
privateWrite: anInteger from: aSequenceableCollection at: startIndex
	" Ensure we're only doing with (writePosition + anInteger) <= cache size "
	(writePosition + anInteger) <= cache size ifFalse: [self error: 'invalid privateWrite. Use the #write: protocol instead.'].

	cache replaceFrom: writePosition + 1 to: writePosition + anInteger with: aSequenceableCollection startingAt: startIndex.
	writePosition := writePosition + anInteger.
	writePosition > dataLength ifTrue: [dataLength := writePosition]! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:36'!
put: anObject

	writePosition < cache size ifFalse: [self growBy: 1].

	cache at: writePosition + 1 put: anObject.
	writePosition := writePosition + 1.
	writePosition > dataLength ifTrue: [dataLength := writePosition].
	^anObject! !

!XTBuffer methodsFor: 'accessing - reading' stamp: 'MartinKobetic 1/11/2011 17:53'!
read: anInteger into: aSequenceableCollection at: startIndex
	| count |
	count := anInteger min: (dataLength - readPosition).
	aSequenceableCollection replaceFrom: startIndex to: startIndex + count - 1 with: cache startingAt: readPosition + 1.
	readPosition := readPosition + count.
	count < anInteger ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise].
	^anInteger! !

!XTBuffer methodsFor: 'accessing - reading' stamp: 'nice 10/9/2010 18:35'!
readPosition
	^readPosition! !

!XTBuffer methodsFor: 'accessing - reading' stamp: 'nice 10/9/2010 18:35'!
readPosition: aPosition
	readPosition := (aPosition max: 0) min: dataLength! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:33'!
readSize
	"The number of elements available to read"

	^dataLength - readPosition! !

!XTBuffer methodsFor: 'accessing - reading' stamp: 'nice 10/9/2010 18:35'!
readSkip: anInteger
	| old delta |
	old := readPosition.
	self readPosition: old + anInteger.
	delta := readPosition - old.
	^delta! !

!XTBuffer methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 21:44'!
recycle

	cache recycle! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:40'!
streamingInsert: anInteger into: aWriteStream
	| count |
	count := (writePosition - readPosition) min: anInteger.
	count <= 0 ifTrue: [^0].
	aWriteStream insert: count from: cache at: readPosition + 1.
	self readSkip: count! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:41'!
streamingInsertInto: aWriteStream
	| count |
	count := writePosition - readPosition.
	count <= 0 ifTrue: [^0].
	self streamingInsert: count into: aWriteStream.
	^count! !

!XTBuffer methodsFor: 'printing' stamp: 'nice 10/9/2010 18:39'!
streamingPrintOn: aStream
	aStream write: self class name;
		space; write: 'data: '; print: dataLength;
		space; write: 'read: '; print: self readPosition;
		space; write: 'write: '; print: self writePosition! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:41'!
streamingWrite: anInteger from: aReadStream
	"Implementing this method would allow the buffer to be treated like a stream sometimes. Currently this is not implemented, make a buffer stream on your buffer instead."
	self error: 'not yet implemented'! !

!XTBuffer methodsFor: 'private' stamp: 'MartinKobetic 12/22/2010 16:40'!
streamingWrite: anInteger into: aWriteStream
	| count |
	count := (writePosition - readPosition) min: anInteger.
	count <= 0 ifTrue: [^0].
	aWriteStream write: count from: cache at: readPosition + 1.
	self readSkip: count.
	^count! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:41'!
streamingWriteFrom: aReadStream
	"Implementing this method would allow the buffer to be treated like a stream sometimes. Currently this is not implemented, make a buffer stream on your buffer instead."
	self error: 'not yet implemented'! !

!XTBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:41'!
streamingWriteInto: aWriteStream
	| count |
	count := writePosition - readPosition.
	count <= 0 ifTrue: [^0].
	self streamingWrite: count into: aWriteStream.
	^count! !

!XTBuffer methodsFor: 'compressing' stamp: 'nice 10/9/2010 18:38'!
trim
	cache := cache copyFrom: readPosition + 1 to: writePosition.
	readPosition := 0.
	dataLength := writePosition := cache size! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:40'!
write: aStreamable

	^aStreamable streamingWriteInto: self! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:37'!
write: anInteger from: aStreamable

	^aStreamable streamingWrite: anInteger into: self! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:40'!
write: anInteger from: aSequenceableCollection at: startIndex

	writePosition + anInteger <= cache size ifFalse: [self growBy: anInteger].
	self privateWrite: anInteger from: aSequenceableCollection at: startIndex.
	^anInteger! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:37'!
writePosition
	^writePosition! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:41'!
writePosition: aPosition
	^writePosition := (aPosition max: 0) min: dataLength! !

!XTBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 18:33'!
writeSize
	"The number of available slots to write in to, which might overwrite elements in dataLength; writeSize >= inactiveSize"

	^(cache size - writePosition + readPosition) min: cache size! !

!XTBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:37'!
writeSkip: anInteger
	| old delta |
	old := writePosition.
	self writePosition: old + anInteger.
	delta := writePosition - old.
	^delta! !

Object subclass: #XTReadStream
	instanceVariableNames: 'source'
	classVariableNames: ''
	poolDictionaries: 'XTPools'
	category: 'Xtreams-Core'!

!XTReadStream commentStamp: '<historical>' prior: 0!
Abstract superclass of all read streams, defines the API.

Read streams are created by sending #reading to a concrete resource (a.k.a terminal), such as a Collection, SocketAccessor, Filename, etc.

	'testing' reading rest

Transform read streams are created through one of the messages in the 'transforming' protocol sent to other read streams.

	('testing' reading collecting: #asUppercase) rest

Subclasses must implement the following messages:
	#read:into:at:
	#contentsSpecies

Instance Variables
	source	<Object> a read stream or "terminal" producing the elements

!

XTReadStream subclass: #XTPositionReadStream
	instanceVariableNames: 'buffer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Core'!

!XTPositionReadStream commentStamp: 'nice 10/9/2010 21:44' prior: 0!
Wraps a non-positionable stream and provides positioning capability by buffering read elements. Buffering strategy can be configured via different Buffer classes.

Instance Variables
	buffer	<Buffer> holds read elements

!

!XTPositionReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
++ anInteger
	| skipped |
	anInteger < 0 ifTrue: [ ^self -- anInteger negated ].
	skipped := buffer readSkip: anInteger.
	skipped = anInteger ifTrue: [^anInteger].
	[super ++ (anInteger - skipped)] on: Incomplete do: [:incomplete |
		(Incomplete count: incomplete count + skipped) raise].
	^anInteger! !

!XTPositionReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
-- anInteger
	| skipped |
	anInteger < 0 ifTrue: [ ^self ++ anInteger negated ].
	skipped := (buffer readSkip: anInteger negated) negated.
	skipped = anInteger ifTrue: [^anInteger].
	(Incomplete count: skipped) raise! !

!XTPositionReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:14'!
buffer
	^buffer! !

!XTPositionReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 22:04'!
buffer: aBuffer

	buffer recycle.
	buffer := aBuffer! !

!XTPositionReadStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 22:05'!
close
	super close.
	buffer recycle.
	buffer := nil! !

!XTPositionReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:15'!
contentsSpecies
	^source contentsSpecies! !

!XTPositionReadStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:14'!
get
	| object |
	buffer hasDataToRead ifTrue: [^buffer get].
	object := source get.
	buffer put: object.
	buffer readSkip: 1.
	^object! !

!XTPositionReadStream methodsFor: 'testing' stamp: 'nice 10/9/2010 19:22'!
isPositionable
	^true! !

!XTPositionReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
length
	"advance to the end"
	[ [ self ++ SmallInteger maxVal ] repeat ] on: Incomplete do: [ :ex | ].
	^buffer activeSize! !

!XTPositionReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:19'!
on: aSource
	buffer := XTBuffer new: DefaultBufferSize class: aSource contentsSpecies.
	super on: aSource! !

!XTPositionReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:22'!
position
	^buffer readPosition! !

!XTPositionReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
position: aPosition

	| startPosition delta |
	aPosition < 0 ifTrue: [Incomplete zero raise].
	startPosition := buffer readPosition.
	delta := aPosition - startPosition.
	^[	self ++ delta. aPosition
	] on: Incomplete do: [ :ex |
		(Incomplete count: startPosition + ex count) raise ]! !

!XTPositionReadStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 19:26'!
positioning
	^self! !

!XTPositionReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/11/2011 17:53'!
read: anInteger into: aSequenceableCollection at: startIndex

	^[	buffer read: anInteger into: aSequenceableCollection at: startIndex. anInteger
	] on: Incomplete do: [:exception || bufferCount sourceCount |
		bufferCount := exception count.
		sourceCount := anInteger - bufferCount.
		[source read: sourceCount into: aSequenceableCollection at: startIndex + exception count]
			on: Incomplete do: [:incomplete | sourceCount := incomplete count].
		buffer write: sourceCount from: aSequenceableCollection at: startIndex + bufferCount.
		buffer readSkip: sourceCount.
		(sourceCount + bufferCount) < anInteger ifTrue: [
			(Incomplete on: aSequenceableCollection count: sourceCount + bufferCount at: startIndex) raise].
		anInteger ]! !

!XTPositionReadStream methodsFor: 'printing' stamp: 'nice 10/9/2010 19:21'!
streamingPrintOn: aStream
	super streamingPrintOn: aStream.
	aStream
		write: ' buffered: ';
		print: buffer readSize.
	buffer readSize isZero ifTrue: [^self].
	aStream
		cr; tab;
		print: buffer contentsFuture! !

!XTReadStream classSide methodsFor: 'instance creation' stamp: 'nice 10/9/2010 19:32'!
on: aSource
	^self new on: aSource! !

!XTReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
++ anInteger
	"Seek forward by anInteger elements."
	"	anInteger	<Integer>	the number of elements to go forward by
		^<Integer>	the number of elements skipped
	""
		'Hello' reading ++ 2; rest
	"
	| count cache chunk read |
	anInteger < 0 ifTrue: [ ^self -- anInteger negated ].
	anInteger = 0 ifTrue: [^0].
	count := 0.
	chunk := DefaultBufferSize min: anInteger.
	cache := self contentsSpecies newRecycledDefaultSize.
	[[count < anInteger] whileTrue:
		[read := chunk min: (anInteger - count).
		self read: read into: cache at: 1.
		count := count + read]]
			on: Incomplete do: [ :ex | cache recycle. (Incomplete count: count + ex count) raise].
	cache recycle.
	^anInteger! !

!XTReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 22:24'!
+= anInteger
	"Seek from the start of the stream by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	The number of elements to go forward by."
	"
		'Hello' reading rest; += 2; rest
	"
	^self position: anInteger! !

!XTReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:30'!
-- anInteger
	"Seek backward by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	The number of elements to go back by."
	"
		'hello' reading rest; -- 3; rest
	"
	"Subclasses should reimplement this method if the stream is positionable."
	self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
-= anInteger
	"Seek backwards from the end of the stream by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	number of elements to go back by
		^<Integer>	number of elements actually skipped
	""
		'Hello' reading -= 3; rest
	"
	| available |
	(self isPositionable not and: [anInteger isZero]) ifTrue:
		["If we are attempting to skip to end and the stream is not positionable, skip forward until we hit an incomplete"
		[[self ++ SmallInteger maxVal] repeat] on: Incomplete do: [].
		^0 ].
	available := anInteger min: self length.
	self position: self length - available.
	available = anInteger ifTrue: [^anInteger].
	^(Incomplete count: available) raise! !

!XTReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:31'!
available
	"Return the number of elements available. The stream must be positionable."
	"	^	<Integer>	the number of elements available"
	"
		'Hello' reading ++2; available
	"
	^self length - self position! !

!XTReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:28'!
close
	source close! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:27'!
collect: aBlock  
	"Evaluate aBlock with each of the values of the receiver as the argument. Collect the resulting values into a collection."
	"	aBlock	<BlockClosure>	#collect: style block transforming the elements being read
		^		<Collection>
	""
		(1 to: 10) reading collect: [ :e | e * e ]
	"
	| contents |
	contents := self contentsSpecies new writing.
	self do: [:each | contents put: (aBlock value: each)].
	^contents close; destination! !

!XTReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:29'!
contentsSpecies
	"Returns collection class suitable to hold elements of this stream."
	^self subclassResponsibility! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
detect: aBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. Answer the first element for which aBlock evaluates to true."
	"	aBlock	<BlockClosure>	#detect: style block
		^		<Object>	first element for which aBlock evalutes to true
	"
	^self detect: aBlock ifNone: [NotFoundError raise]! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
detect: aBlock ifNone: exceptionBlock 
	"Evaluate aBlock with each of the receiver's elements as the argument. Answer the first element for which aBlock evaluates to true.
	Evaluate exceptionBlock if no such element is found."
	"	aBlock			<BlockClosure>	#detect: style block
		exceptionBlock	<BlockClosure>	evaluated if there wasn't any element for which aBlock returns true
		^				<Object>	first element for which aBlock evalutes to true
	"
	self do: [:each | (aBlock value: each) ifTrue: [^each]].
	^exceptionBlock value! !

!XTReadStream methodsFor: 'enumerating' stamp: 'MartinKobetic 1/11/2011 17:53'!
do: aBlock
	"Read from the stream and pass the read objects to the #do: block one by one until it reaches the end of stream."
	"	aBlock	<BlockClosure>	#do: style block evaluated with each element
	"
	[[aBlock cull: self get] repeat] on: Incomplete do: []! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
do: elementBlock separatedBy: separatorBlock
	"Evaluate elementBlock for each element in the stream. Between each pair of elements, but not before the first or after the last, evaluate the separatorBlock."
	"	elementBlock	<BlockClosure> evaluated with each element
		separatorBlock	<BlockClosure> evaluated between elements
	"
	| first |
	first := true.
	self do:
		[:elmnt |
		first ifTrue: [first := false]
			ifFalse: [separatorBlock value].
		elementBlock value: elmnt]! !

!XTReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:31'!
explore: aBlock
	" Explore the stream within the block but return to where we started when the block completes. The stream must be positionable."
	"	aBlock	<BlockClosure>	defines the exploration activity
		^		<Object>	result of aBlock"
	"
		'Hello' reading explore: [ :s | s -= 0 ]; rest
	"
	| position |
	position := self position.
	^[aBlock cull: self] ensure: [self position: position]! !

!XTReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 12/30/2010 22:25'!
get
	"Read an object from self.
	If there aren't any elements left in the stream, the Incomplete exception is raised."
	"	^	<Object> an object read from this stream
	"
	| cache object |
	cache := self contentsSpecies newRecycled: 1.
	self read: 1 into: cache at: 1.
	object := cache first.
	cache recycle.
	^object! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
groupedBy: aBlock
	"Return a dictionary whose keys are the result of evaluating aBlock for all elements in	 the stream, and the value for each key is a collection of elements that evaluated to that key.
	""	aBlock	<BlockClosure>	#collect: style block evaluated with each element
		^		<Dictionary>		keys are results of aBlock for all elements, values are corresponding elements that evaluated to the key
	""
	     #(1 2 3 4 5) reading groupedBy: [:each | each odd]
	"
	| result |
	result := Dictionary new.
	self do:
		[:each | | key collection |
		key := aBlock value: each.
		collection := result at: key ifAbsentPut: [OrderedCollection new].
		collection add: each].
	self species ~~ OrderedCollection ifTrue:
		["Convert the result collections to be the right type.
		  Note that it should be safe to modify the dictionary
		  while iterating because we only replace values for existing keys"
		result keysAndValuesDo:
			[:key :value | result at: key put: (self contentsSpecies withAll: value)]].

	^result! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
inject: initialValue into: binaryBlock 
	"Accumulate a running value associated with evaluation of binaryBlock with the each element and the result of evaluation for previous element. 
	The initialValue serves as the previous result for evaluation of the first element."
	"	initialValue	<Object>	servers as the previous result for evaluation of the first element.
		binaryBlock	<BlockClosure>	evaluated with each element and result of evaluation of previous element
	""
		(1 to: 10) reading inject: 0 into: [:subTotal :next | subTotal + next].
	"
	| nextValue |
	nextValue := initialValue.
	self do: [:each | nextValue := binaryBlock value: nextValue value: each].
	^nextValue! !

!XTReadStream methodsFor: 'testing' stamp: 'nice 10/9/2010 19:31'!
isPositionable
	"Can this stream be positioned. Positionable streams support additional API: #position, #position:, ++, --, ..."

	^false! !

!XTReadStream methodsFor: 'testing' stamp: 'nice 10/9/2010 19:31'!
isReadable
	^true! !

!XTReadStream methodsFor: 'testing' stamp: 'nice 10/9/2010 19:32'!
isWritable
	^false! !

!XTReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:31'!
length
	"Return total length of the stream. The stream must be positionable."
	"	^	<Integer>	the total number of elements in the stream. (position + available)"
	"
		'Hello' reading read: 2; length
	"
	"Subclasses should reimplement this method if the stream is positionable."
	^self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTReadStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:29'!
on: aSource
	source := aSource! !

!XTReadStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:31'!
position
	"Return current position of the stream. The stream must be positionable."
	"	^	<Integer>	the current position in the stream."
	"
		'Hello' reading read: 2; position
	"
	"Subclasses should reimplement this method if the stream is positionable."
	^self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTReadStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 22:26'!
position: anInteger
	"Change position of the stream to anInteger. The stream must be positionable."
	"	anInteger	<Integer>	the position to set the stream at.
		^<Integer>	the position the stream was set to
	""
		'Hello' reading position: 2; rest
	"
	"Subclasses should reimplement this method if the stream is positionable."
	self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTReadStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 19:32'!
positioning
	"If necessary add positioning layer. Note that positiong layer employs buffering to implement the positioning ability. The default buffering strategy will grow the buffer up to the full size of the underlying stream if not released. Consequently other Buffer types might be more suitable for specific circumstances, e.g. if only last n elements need to be buffered, a fixed size RingBuffer can be substitued with #buffer: accessor."
	"	^	<ReadStream>	a positionable read stream
	""
		[ Time now ] reading positioning ++ 3; -- 2; get
	"
	^self isPositionable
		ifTrue:	[self]
		ifFalse:	[XTPositionReadStream on: self]! !

!XTReadStream methodsFor: 'printing' stamp: 'nice 10/9/2010 19:29'!
printOn: aStream
	| stream |
	stream := String new writing.
	self streamingPrintOn: stream.
	aStream nextPutAll: stream conclusion.
	aStream cr.
	source printOn: aStream! !

!XTReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:10'!
read: anInteger
	"Read anInteger's worth of elements from self and return them in a collection.
	If full anInteger number of elements cannot be read from the source, the Incomplete exception is raised."
	"	anInteger	<Integer>	the number of elements to read
 	 	^<SequenceableCollection>	a new collection containing anIntegers worth of elements
	"
	| newCollection |
	newCollection := self contentsSpecies withSize: anInteger.
	self read: anInteger into: newCollection at: 1.
	^newCollection! !

!XTReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:10'!
read: anInteger into: aSequenceableCollection
	"Read anInteger's worth of elements into aSequenceableCollection starting from index 1.
	If full anInteger number of elements cannot be read from the source, the Incomplete exception is raised."
	"	anInteger	<Integer>	the number of elements to read
	 	aSequenceableCollection	<SequenceableCollection>	the destination to read into
		^<Integer>	number of elements read
	"
	^self read: anInteger into: aSequenceableCollection at: 1! !

!XTReadStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:11'!
read: anInteger into: aSequenceableCollection at: startIndex
	"Read anInteger's worth of elements into aSequenceableCollection starting at startIndex.
	If full anInteger number of elements cannot be read from the source, the Incomplete exception is raised."
	"	anInteger	<Integer>	the number of elements to read
	 	aSequenceableCollection	<SequenceableCollection>	the destination to read into
		startIndex	<Integer>	the index into aSequenceableCollection to start writing to
		^<Integer>	number of elements read
	"
	self subclassResponsibility
	"Subclasses must implement this message. All effort must be taken to read anInteger, blocking if necessary. "! !

!XTReadStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 19:32'!
reading
	"Return a readable stream. This is to make the method polymorphic between read streams and terminals."
	"	^	<ReadStream>
	"
	^self! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
reject: aBlock 
	"Evaluate aBlock with each element, collect into a collection only those elements for which aBlock evaluates to false."
	"	aBlock	<BlockClosure>	#reject: style block used to filter the elements
		^		<Collection>		all elements that evaluate to false 
	""
		(1 to: 10) reading reject: [ :e | e odd ]
	"
	^self select: [:element | (aBlock value: element) == false]! !

!XTReadStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:27'!
rest
	"Read all remaining elements from self."
	"	^		<SequenceableCollection>	a new collection of elements read from this stream
	"

	"Implementation note: Originally we sent #close then #destination, but sending #close will perform a #become:
		and our constructed write stream is going to go away as soon as we leave this method, so we're better off
		using #contents instead"

	^self contentsSpecies new writing
		write: self;
		contents! !

!XTReadStream methodsFor: 'enumerating' stamp: 'nice 10/9/2010 19:28'!
select: aBlock 
	"Evaluate aBlock with each element, collect into a collection only those elements for which aBlock evaluates to true."
	"	aBlock	<BlockClosure>	#select: style block used to filter the elements
		^		<Collection>		all elements that evaluate to true 
	""
		(1 to: 10) reading select: [ :e | e odd ]
	"
	| contents |
	contents := self contentsSpecies new writing.
	self do: [:each | (aBlock value: each) ifTrue: [contents put: each]].
	^contents close; destination! !

!XTReadStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:27'!
source
	"Return the source of this stream."

	^source! !

!XTReadStream methodsFor: 'private' stamp: 'nice 10/9/2010 19:29'!
streamingInsert: anInteger into: aWriteStream
	aWriteStream streamingInsert: anInteger from: self! !

!XTReadStream methodsFor: 'private' stamp: 'nice 10/9/2010 19:30'!
streamingInsertInto: aWriteStream
	^aWriteStream streamingInsertFrom: self! !

!XTReadStream methodsFor: 'printing' stamp: 'nice 10/9/2010 19:29'!
streamingPrintOn: aStream
	aStream write: self class name! !

!XTReadStream methodsFor: 'private' stamp: 'MartinKobetic 12/30/2010 22:27'!
streamingWrite: anInteger into: aWriteStream
	^aWriteStream streamingWrite: anInteger from: self! !

!XTReadStream methodsFor: 'private' stamp: 'nice 10/9/2010 19:30'!
streamingWriteInto: aWriteStream
	^aWriteStream streamingWriteFrom: self! !

!XTReadStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:27'!
terminal
	"Return the object at the bottom of the stream."
	"	^<Collection | Buffer | IOAccessor | BlockClosure>
	"
	^(source isKindOf: XTReadStream)
		ifTrue: [ source terminal ]
		ifFalse: [ source ]! !

XTBuffer subclass: #XTRingBuffer
	instanceVariableNames: 'dataPosition'
	classVariableNames: ''
	poolDictionaries: 'XTPools'
	category: 'Xtreams-Core'!

!XTRingBuffer commentStamp: 'nice 10/9/2010 21:45' prior: 0!
RingBuffer is a specialized buffer that has a fixed sized cache. The cache is treated as ring/circle, such that when we get to the end of the cache, we continue reading and writing data from the beginning of the cache. If you write more data than you have written, the unread data is lost. The readPosition and writePosition are always relative to the dataPosition, which is a hidden implementation detail.

Instance Variables
	dataPosition	<ArithmeticValue>	the position in our cache where our data starts

!

XTRingBuffer subclass: #XTElasticBuffer
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Core'!

!XTElasticBuffer commentStamp: '<historical>' prior: 0!
ElasticBuffer is a variation on a ring buffer that will grow if it has no room to write the data in to its cache. It will split the cache and insert new elements if necessary, otherwise it will wrap around like a regular ring buffer. You have a slightly more liberal 'rewind' policy than the RingBuffer such that if you read data, then write data, it will only overwrite the read data if the write data can fit in to that space, otherwise it will grow the cache.

!

!XTElasticBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:51'!
clearCache
	cache := cache copyEmpty: 0! !

!XTElasticBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 18:56'!
growBy: anInteger
	| split replacement growth |
	split := dataPosition + writePosition.
	cache size isZero ifFalse: [split := split \\ cache size].
	split isZero ifTrue: [split := cache size].
	growth := cache size + anInteger + anInteger.
	replacement := cache copyEmpty: cache size + growth.
	replacement replaceFrom: 1 to: split with: cache startingAt: 1.
	replacement replaceFrom: split + growth + 1 to: replacement size with: cache startingAt: split + 1.
	cache := replacement.
	split <= dataPosition ifTrue: [dataPosition := dataPosition + growth]! !

!XTElasticBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 18:56'!
hasFixedWriteSpace
	^false! !

!XTElasticBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 18:57'!
hasSpaceToWrite
	^true! !

!XTElasticBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 18:47'!
put: anObject

	self writeSize < 1 ifTrue: [self growBy: 1].
	^super put: anObject! !

!XTElasticBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/22/2010 16:41'!
write: anInteger from: aSequenceableCollection at: startIndex

	anInteger <= self writeSize ifFalse: [self growBy: anInteger - self writeSize].
	self privateWrite: anInteger from: aSequenceableCollection at: startIndex.
	^anInteger! !

!XTRingBuffer methodsFor: 'compressing' stamp: 'nice 10/9/2010 19:46'!
clear
	super clear.
	dataPosition := 0! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:46'!
clearCache! !

!XTRingBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:44'!
contentsFuture
	| buffer read tail |
	cache isEmpty ifTrue: [^self contentsSpecies new].
	read := (dataPosition + readPosition) \\ cache size.
	tail := (cache size - read) min: self readSize.

	buffer := self contentsSpecies withSize: self readSize.
	buffer replaceFrom: 1 to: tail with: cache startingAt: read + 1.
	buffer replaceFrom: tail + 1 to: self readSize - tail with: cache startingAt: 1.
	^buffer! !

!XTRingBuffer methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:44'!
contentsPast
	| buffer tail |
	cache isEmpty ifTrue: [^self contentsSpecies new].
	tail := (cache size - dataPosition) min: self writePosition.

	buffer := self contentsSpecies withSize: self writePosition.
	buffer replaceFrom: 1 to: tail with: cache startingAt: dataPosition + 1.
	buffer replaceFrom: tail + 1 to: buffer size with: cache startingAt: 1.
	^buffer! !

!XTRingBuffer methodsFor: 'accessing - reading' stamp: 'MartinKobetic 1/11/2011 17:53'!
get
	| read |
	dataLength = readPosition ifTrue: [Incomplete zero raise].
	read := (dataPosition + readPosition) \\ cache size.
	readPosition := readPosition + 1.
	^cache at: read + 1! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:46'!
growBy: anInteger
	self shouldNotImplement! !

!XTRingBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 19:47'!
hasFixedWriteSpace
	^true! !

!XTRingBuffer methodsFor: 'testing' stamp: 'nice 10/9/2010 19:47'!
hasSpaceToWrite
	^self writeSize isZero not! !

!XTRingBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/30/2010 22:31'!
insert: anInteger from: aSequenceableCollection at: startIndex

	| tail write temp length position |
	length := dataLength - writePosition.
	write := (dataPosition + writePosition + 1) \\ cache size.
	tail := (cache size - write) min: length.
	temp := self contentsSpecies newRecycled: (length max: DefaultBufferSize).
	temp replaceFrom: 1 to: tail with: cache startingAt: write.
	temp replaceFrom: tail + 1 to: length with: cache startingAt: 1.
	dataLength := dataLength - length.
	self write: anInteger from: aSequenceableCollection at: startIndex.
	position := self writePosition.
	self write: length from: temp.
	self writePosition: position. 
	temp recycle.
	^anInteger! !

!XTRingBuffer methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:46'!
on: aSequenceableCollection
	super on: aSequenceableCollection.
	dataPosition := 0! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:47'!
privateWrapDataPosition: anInteger
"	inflatedLength <= cache size --- do nothing
	inflatedLength >   cache size --- move dataPosition and readPosition
"
	| inflatedLength overflow |

	inflatedLength := writePosition + anInteger.
	inflatedLength <= cache size ifTrue: [^self].
	overflow := inflatedLength - cache size.
	readPosition := (readPosition - overflow) max: 0.
	dataPosition := (dataPosition + overflow) \\ cache size! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:47'!
privateWrite: anInteger from: aSequenceableCollection at: startIndex
	| write tail |

	anInteger > 0 ifFalse: [^self].
	" Ensure we're only doing with anInteger <= cache size "
	anInteger > cache size ifTrue: [self error: 'invalid privateWrite. Use the #write: protocol instead.'].

	write := (dataPosition + writePosition) \\ cache size.
	tail := (cache size - write) min: anInteger.

	cache replaceFrom: write + 1 to: write + tail with: aSequenceableCollection startingAt: startIndex.
	cache replaceFrom: 1 to: anInteger - tail with: aSequenceableCollection startingAt: startIndex + tail.

	self privateWrapDataPosition: anInteger.
	writePosition := (writePosition + anInteger) min: cache size.
	writePosition > dataLength ifTrue: [dataLength := writePosition]! !

!XTRingBuffer methodsFor: 'accessing - writing' stamp: 'nice 10/9/2010 19:45'!
put: anObject
	| write |

	write := (dataPosition + writePosition) \\ cache size.
	cache at: write + 1 put: anObject.

	self privateWrapDataPosition: 1.
	writePosition := (writePosition + 1) min: cache size.
	writePosition > dataLength ifTrue: [dataLength := writePosition].
	^anObject! !

!XTRingBuffer methodsFor: 'accessing - reading' stamp: 'MartinKobetic 1/11/2011 17:53'!
read: anInteger into: aSequenceableCollection at: startIndex
	| count read tail |

	count := anInteger min: (dataLength - readPosition).
	read := (dataPosition + readPosition) \\ cache size.
	tail := (cache size - read) min: count.

	aSequenceableCollection replaceFrom: startIndex to: startIndex + tail - 1 with: cache startingAt: read + 1.
	aSequenceableCollection replaceFrom: startIndex + tail to: startIndex + count - 1 with: cache startingAt: 1.
	readPosition := readPosition + count.

	count < anInteger ifTrue: [(Incomplete on: aSequenceableCollection count: count at: startIndex) raise].
	^anInteger! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:47'!
streamingInsert: anInteger into: aWriteStream

	| read amount tail head |
	amount := (dataLength - readPosition) min: anInteger.
	read := (dataPosition + readPosition) \\ cache size.
	tail := cache size - read min: amount.
	head := amount - tail.
	aWriteStream insert: tail from: cache at: read + 1.
	head <= 0 ifTrue: [^self readSkip: tail].
	aWriteStream insert: head from: cache at: 1.
	self readSkip: tail + head! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:47'!
streamingInsertInto: aWriteStream

	| count |
	count := dataLength - readPosition.
	self streamingInsert: count into: aWriteStream.
	^count! !

!XTRingBuffer methodsFor: 'private' stamp: 'MartinKobetic 12/30/2010 22:32'!
streamingWrite: anInteger into: aWriteStream

	| read amount tail head |
	amount := (dataLength - readPosition) min: anInteger.
	read := (dataPosition + readPosition) \\ cache size.
	tail := cache size - read min: amount.
	head := amount - tail.
	aWriteStream write: tail from: cache at: read + 1.
	head <= 0 ifTrue: [self readSkip: tail. ^anInteger].
	aWriteStream write: head from: cache at: 1.
	self readSkip: tail + head.
	^anInteger! !

!XTRingBuffer methodsFor: 'private' stamp: 'nice 10/9/2010 19:47'!
streamingWriteInto: aWriteStream

	| count |
	count := dataLength - readPosition.
	self streamingWrite: count into: aWriteStream.
	^count! !

!XTRingBuffer methodsFor: 'compressing' stamp: 'nice 10/9/2010 19:46'!
trim
	| read write |
	read := (dataPosition + readPosition) \\ cache size.
	write := (dataPosition + writePosition) \\ cache size.
	read <= write
		ifTrue: [cache := cache copyFrom: read + 1 to: write + 1]
		ifFalse:
			[ | replacement readSize |
			readSize := cache size - read.
			replacement := cache copyEmpty: write + readSize.
			replacement replaceFrom: 1 to: readSize with: cache startingAt: read + 1.
			replacement replaceFrom: readSize + 1 to: replacement size with: cache startingAt: write + 1.
			cache := replacement].
	dataPosition := readPosition := 0.
	dataLength := writePosition := cache size! !

!XTRingBuffer methodsFor: 'accessing - writing' stamp: 'MartinKobetic 12/30/2010 22:33'!
write: anInteger from: aSequenceableCollection at: startIndex

	anInteger < cache size ifTrue:  [self privateWrite: anInteger from: aSequenceableCollection at: startIndex. ^anInteger].

	" Do the quick store, since we're a ring "
	dataPosition := readPosition := 0.
	dataLength := writePosition := cache size.
	cache replaceFrom: 1 to: cache size with: aSequenceableCollection startingAt: startIndex + anInteger - cache size.
	^anInteger! !

Object subclass: #XTWriteStream
	instanceVariableNames: 'destination'
	classVariableNames: 'Backspace Bell CarriageReturn Delete DoubleQuote Escape FormFeed LineFeed Quote Space Tab VerticalTab'
	poolDictionaries: 'XTPools'
	category: 'Xtreams-Core'!

!XTWriteStream commentStamp: '<historical>' prior: 0!
Abstract superclass of all write streams; defines the API.

Write streams are created by sending #writing to a concrete resource (a.k.a terminal), such as a Collection, SocketAccessor, Filename, etc.

	String new writing write: 'testing'; close; terminal

Transform write streams are created through one of the messages in the 'transforming' protocol sent to other write streams.

	(String new writing collecting: #asUppercase) write: 'testing'; close; terminal

Subclasses must implement the following messages:
	#read:into:at:
	#contentsSpecies

Instance Variables
	destination	<Object> a stream or "terminal" consuming written elements

Shared Variables
	Backspace	<Character>
	Bell	<Character>
	CarriageReturn	<Character>
	Delete	<Character>
	DoubleQuote	<Character>
	Escape	<Character>
	FormFeed	<Character>
	LineFeed	<Character>
	Quote	<Character>
	Space	<Character>
	Tab	<Character>
	VerticalTab	<Character>

!

XTWriteStream subclass: #XTPositionWriteStream
	instanceVariableNames: 'buffer'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xtreams-Core'!

!XTPositionWriteStream commentStamp: 'nice 10/9/2010 21:44' prior: 0!
Wraps a non-positionable stream and provides positioning capability by buffering written elements. Buffering strategy can be configured via different Buffer classes.

Instance Variables
	buffer	<Buffer> holds written elements

!

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
++ anInteger
	| skipped |
	anInteger < 0 ifTrue: [ ^self -- anInteger negated ].
	skipped := buffer writeSkip: anInteger.
	skipped = anInteger ifTrue: [^anInteger].
	[destination ++ (anInteger - skipped)] on: Incomplete do: [:incomplete |
		(Incomplete count: incomplete count + skipped) raise].
	^anInteger! !

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
-- anInteger
	| skipped |
	anInteger < 0 ifTrue: [ ^self ++ anInteger negated ].
	skipped := (buffer writeSkip: anInteger negated) negated.
	skipped = anInteger ifTrue: [^anInteger].
	(Incomplete count: skipped) raise! !

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:25'!
available
	^buffer writeSize! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:23'!
buffer

	^buffer! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 22:14'!
buffer: aBuffer

	buffer recycle.
	buffer := aBuffer! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'MartinKobetic 12/30/2010 22:14'!
close
	super close.
	buffer recycle.
	buffer := nil! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:23'!
contentsSpecies
	^destination contentsSpecies! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:24'!
flush
	buffer ifNil: [^self].
	destination write: buffer.
	buffer clear! !

!XTPositionWriteStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:23'!
insert: anInteger from: aSequenceableCollection at: startIndex
	buffer insert: anInteger from: aSequenceableCollection at: startIndex! !

!XTPositionWriteStream methodsFor: 'testing' stamp: 'nice 10/9/2010 19:25'!
isPositionable
	^true! !

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:25'!
length
	^buffer activeSize! !

!XTPositionWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 19:24'!
on: aDestination
	buffer := XTBuffer new: DefaultBufferSize class: aDestination contentsSpecies.
	super on: aDestination! !

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 19:25'!
position
	^buffer writePosition! !

!XTPositionWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
position: aPosition
	| startPosition delta |
	aPosition < 0 ifTrue: [Incomplete zero raise].
	startPosition := buffer writePosition.
	delta := aPosition - startPosition.
	^[	self ++ delta. aPosition
	] on: Incomplete do: [ :ex |
		(Incomplete count: startPosition + ex count) raise ]! !

!XTPositionWriteStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 19:25'!
positioning
	^self! !

!XTPositionWriteStream methodsFor: 'printing' stamp: 'nice 10/9/2010 19:24'!
streamingPrintOn: aStream
	super streamingPrintOn: aStream.
	aStream
		write: ' buffered: ';
		print: buffer writeSize.
	buffer writeSize isZero ifTrue: [^self].
	aStream
		cr; tab;
		print: buffer contentsPast! !

!XTPositionWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 12/30/2010 22:17'!
write: anInteger from: aSequenceableCollection at: startIndex

	| toFlush bufferFlush |
	(buffer hasFixedWriteSpace not or: [buffer writeSize >= anInteger ]) ifTrue: [
		buffer write: anInteger from: aSequenceableCollection at: startIndex.
		^anInteger ].
	toFlush := buffer readSize + anInteger - buffer cacheSize.
	bufferFlush := toFlush min: buffer readSize.
	destination write: bufferFlush from: buffer.
	(toFlush > bufferFlush) ifTrue: [
		toFlush := toFlush - bufferFlush.
		destination write: toFlush from: aSequenceableCollection at: startIndex ].
	buffer write: anInteger - toFlush from: aSequenceableCollection at: startIndex + toFlush.
	^anInteger! !

!XTWriteStream classSide methodsFor: 'class initialization' stamp: 'MartinKobetic 12/30/2010 22:52'!
initialize
	Backspace := String with: Character backspace.
	Bell := String with: (Character value: 7).
	CarriageReturn := String with: Character cr.
	Delete := String with: (Character value: 127).
	DoubleQuote := String with: $".
	Escape := String with: (Character value: 27).
	FormFeed := String with: Character newPage.
	LineFeed := String with: Character lf.
	Quote := String with: $'.
	Space := String with: Character space.
	Tab := String with: Character tab.
	VerticalTab := String with: (Character value: 11)! !

!XTWriteStream classSide methodsFor: 'instance creation' stamp: 'nice 10/9/2010 19:57'!
on: aDestination
	^self new on: aDestination! !

!XTWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
++ anInteger
	"Seek forward by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	the number of elements to go forward by.
		^<Integer>	the number of elements actually skipped
	"
	"
		'Hello Would' copy writing ++ 6; write: 'World'; close; destination
	"
	"Subclasses should reimplement this method if the stream is positionable."
	self isPositionable
		ifFalse:	[Incomplete zero raise]
		ifTrue:	[self subclassResponsibility]! !

!XTWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 12/30/2010 22:38'!
+= anInteger
	"Seek from the start of the stream by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	The number of elements to go forward by."
	"
		String new writing write: 'Hello Would'; += 6; write: 'World'; close; destination
	"
	^self position: anInteger! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:02'!
-- anInteger
	"Seek backward by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	The number of elements to go back by."
	"
		String new writing write: 'helio'; -- 2; write: 'lo'; close; destination
	"
	"Subclasses should reimplement this method if the stream is positionable."
	self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTWriteStream methodsFor: 'seeking' stamp: 'MartinKobetic 1/11/2011 17:53'!
-= anInteger
	"Seek backwards from the end of the stream by anInteger elements. The stream must be positionable."
	"	anInteger	<Integer>	The number of elements to go back by.
		^<Integer>	the number of elements actually skipped"
	"
		'Hello Would' copy writing -= 3; write: 'rld'; close; terminal
	"
	| available |
	available := anInteger min: self length.
	self position: self length - available.
	available = anInteger ifTrue: [ ^anInteger ].
	^(Incomplete count: available) raise! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:02'!
available
	"Return the number of elements from the current position to the end of the stream. The stream must be positionable."
	"	^	<Integer>	the number of elements available"
	"
		String new writing write: 'Hello World'; -- 5; available
	"
	^self length - self position! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:57'!
backspace
	self write: Backspace! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:58'!
bell
	self write: Bell! !

!XTWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 20:00'!
close
	"Close the destination from any more writes."

	self flush.
	destination close! !

!XTWriteStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:48'!
conclusion
	"Close the stream and return the object at the bottom of the stream."
	"	^<Collection | Buffer | IOAccessor | BlockClosure> "
	self close.
	^self terminal! !

!XTWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 20:00'!
contentsSpecies
	"The class of collection that is able to hold the kind of elements that this stream consumes."
	"	^	<Class>	collection class
	"
	^self subclassResponsibility! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:58'!
cr
	self write: CarriageReturn! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:58'!
delete
	self write: Delete! !

!XTWriteStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:48'!
destination

	^destination! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:58'!
escape
	self write: Escape! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:03'!
explore: aBlock
	" Explore the stream within the block but return to where we started when the block completes. The stream must be positionable."
	"	aBlock	<BlockClosure>	defines the exploration activity
		^		<Object>	result of aBlock"
	"
		String new writing explore: [ :s | s write: 'Hello' ]; write: 'World'; close; destination
	"
	| position |
	position := self position.
	^[aBlock cull: self] ensure: [self position: position]! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:58'!
ff
	self write: FormFeed! !

!XTWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 20:00'!
flush
	"Make sure all the previously written elements are pushed down into the destination."
	destination flush! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:09'!
insert: aStreamable
	"Insert aStreamable into self at current position."
	"	aStreamable	<SequenceableCollection | ReadStream | Buffer>	the source to write in to the destination
	 	^<Integer>	the number of elements written to the destination"
	"
		' World!!' copy writing insert: 'Hello' reading; -= 0; close; destination
	"
	^aStreamable streamingInsertInto: self! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 15:52'!
insert: anInteger from: aStreamable
	"Insert anIntegers worth of elements from aStreamable into self at current position."
	"	anInteger	<Integer>	the number of elements to insert
	 	aStreamable	<ReadStream | SequenceableCollection | Buffer >	the source to write into the destination
	 	startIndex	<Integer>	the index into aSequenceableCollection to start writing from
		^<Integer>	number of elements inserted
	""
		' World!!' copy writing insert: 5 from: 'Hello Underworld!!' reading; -= 0; close; destination
	"
	aStreamable streamingInsert: anInteger into: self.
	^anInteger! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 15:54'!
insert: anInteger from: aSequenceableCollection at: startIndex
	"Insert anIntegers worth of elements from aSequenceableCollection starting at startIndex into self at current position."
	"	anInteger	<Integer>	the number of elements to insert
	 	aStreamable	<SequenceableCollection>	the source to write into the destination
	 	startIndex	<Integer>	the index into aSequenceableCollection to start writing from
		^<Integer>	number of elements inserted
	""
		' World!!' copy writing insert: 5 from: 'Hello' at: 1; -= 0; close; destination
	"
	self write: anInteger from: aSequenceableCollection at: startIndex.
	^anInteger! !

!XTWriteStream methodsFor: 'testing' stamp: 'nice 10/9/2010 20:03'!
isPositionable
	"Can this stream be positioned. Positionable streams come with extra API: #position, #position:, etc."

	^false! !

!XTWriteStream methodsFor: 'testing' stamp: 'nice 10/9/2010 20:03'!
isReadable
	^false! !

!XTWriteStream methodsFor: 'testing' stamp: 'nice 10/9/2010 20:03'!
isWritable
	^true! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:03'!
length
	"Return total length of the stream. The stream must be positionable."
	"	^	<Integer>	the total number of elements in the stream. (position + available)"
	"
		'Hello World' copy writing ++ 5; length
	"
	"Subclasses should reimplement this method if the stream is positionable."
	^self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:59'!
lf
	self write: LineFeed! !

!XTWriteStream methodsFor: 'initialize-release' stamp: 'nice 10/9/2010 20:00'!
on: aDestination
	destination := aDestination! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:03'!
position
	"Return current position of the stream. The stream must be positionable."
	"	^	<Integer>	current position of the stream."
	"
		'Hello World' copy writing -= 5; position
	"
	"Subclasses should reimplement this method if the stream is positionable."
	^self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTWriteStream methodsFor: 'seeking' stamp: 'nice 10/9/2010 20:03'!
position: anInteger
	"Change current position of the stream to anInteger. The stream must be positionable."
	"	anInteger	<Integer>	current position of the stream."
	"
		'Hello Would' copy writing position: 6; write: 'World'; close; destination
	"
	"Subclasses should reimplement this method if the stream is positionable."
	self isPositionable
		ifFalse:	[self error: 'This stream is not positionable.']
		ifTrue:	[self subclassResponsibility]! !

!XTWriteStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 20:04'!
positioning
	"If necessary add positioning layer. Note that positiong layer employs buffering to implement the positioning ability. The default buffering strategy will grow the buffer up to the full size of the underlying stream if not released. Consequently other Buffer types might be more suitable for specific circumstances, e.g. if only last n elements need to be buffered, a fixed size RingBuffer can be substitued with #buffer: accessor."
	"	^	<WriteStream>	a positionable read stream
	""
		[ :x | Transcript space; print: x ] writing positioning write: (1 to: 10); -- 5; write: (11 to: 15); close
	"
	^self isPositionable
		ifTrue:	[self]
		ifFalse:	[XTPositionWriteStream on: self]! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:59'!
print: anObject
	anObject streamingPrintOn: self! !

!XTWriteStream methodsFor: 'printing' stamp: 'nice 10/9/2010 20:01'!
printOn: aStream
	| stream |
	stream := String new writing.
	self streamingPrintOn: stream.
	aStream nextPutAll: stream conclusion.
	aStream cr.
	destination printOn: aStream.! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 12/30/2010 22:41'!
put: anObject
	"Write anObject into self."
	"	anObject		<Object>	the object to write in to the destination
	 	^				<Object>	the object that was written to the destination
	""
		String new writing put: $h; close; destination
	"
	| cache |
	cache := self contentsSpecies newRecycled: 1.
	cache at: 1 put: anObject.
	self write: 1 from: cache at: 1.
	cache recycle.
	^anObject! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 19:59'!
q
	self write: Quote! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
qq
	self write: DoubleQuote! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
space
	self write: Space! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
space: anInteger
	anInteger timesRepeat: [self space]! !

!XTWriteStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
streamingInsert: anInteger from: aReadStream
	| cache count |
	cache := self contentsSpecies newRecycledAtLeast: anInteger.
	count := [aReadStream read: anInteger into: cache at: 1. anInteger] on: Incomplete do: [ :ex | ex count ].
	self insert: count from: cache at: 1.
	cache recycle.
	count < anInteger ifTrue: [(Incomplete count: count) raise]! !

!XTWriteStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
streamingInsertFrom: aReadStream
	| count cache |
	count := 0.
	cache := self contentsSpecies newRecycledDefaultSize.
	[[aReadStream read: cache size into: cache at: 1] on: Incomplete do: [:exception |
		self insert: exception.
		cache recycle.
		^count + exception count].
	self insert: cache size from: cache at: 1.
	count := count + cache size] repeat! !

!XTWriteStream methodsFor: 'printing' stamp: 'nice 10/9/2010 20:01'!
streamingPrintOn: aStream
	aStream write: self class name! !

!XTWriteStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
streamingWrite: anInteger from: aReadStream
	| cache toDo continue amount |
	cache := self contentsSpecies newRecycledDefaultSize.
	toDo := anInteger. continue := true.
	[ continue and: [ toDo > 0 ] ] whileTrue: [
		amount := [ aReadStream read: (cache size min: toDo) into: cache at: 1 ] on: Incomplete do: [ :ex | continue := false. ex count ].
		self write: amount from: cache at: 1.
		toDo := toDo - amount ].
	cache recycle.
	toDo > 0 ifTrue: [(Incomplete count: anInteger - toDo) raise].
	^anInteger! !

!XTWriteStream methodsFor: 'private' stamp: 'MartinKobetic 1/11/2011 17:53'!
streamingWriteFrom: aReadStream
	| count cache |
	count := 0.
	cache := self contentsSpecies newRecycledDefaultSize.
	[[aReadStream read: cache size into: cache at: 1] on: Incomplete do: [:exception |
		self write: exception.
		cache recycle.
		^count + exception count].
	self write: cache size from: cache at: 1.
	count := count + cache size] repeat! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
tab
	self write: Tab! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
tab: anInteger
	anInteger timesRepeat: [self tab]! !

!XTWriteStream methodsFor: 'accessing' stamp: 'nice 10/9/2010 19:49'!
terminal
	"Return the object at the bottom of the stream."
	"	^<Collection | Buffer | IOAccessor | BlockClosure>
	"
	^(destination isKindOf: XTWriteStream)
		ifTrue: [ destination terminal ]
		ifFalse: [ destination ]! !

!XTWriteStream methodsFor: 'characters' stamp: 'nice 10/9/2010 20:00'!
vtab
	self write: VerticalTab! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:06'!
write: aStreamable
	"Write aStreamable into self."
	"	aStreamable	<SequenceableCollection | ReadStream | Buffer>	the source to write in to the destination
	 	^<Integer>	the number of elements written to the destination"
	"
		String new writing write: 'Hello' reading; close; destination
	"
	^aStreamable streamingWriteInto: self! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:05'!
write: anInteger from: aStreamable
	"Write anInteger's worth of elements from aStreamable into self."
	"	anInteger	<Integer>	the number of elements to write
	 	aStreamable	<SequenceableCollection | ReadStream | Buffer>	the source to write in to the destination
		^<Integer>	number of elements written
	""
		String new writing write: 3 from: 'Hello' reading; close; destination
	"
	^aStreamable streamingWrite: anInteger into: self! !

!XTWriteStream methodsFor: 'accessing' stamp: 'MartinKobetic 1/1/2011 16:06'!
write: anInteger from: aSequenceableCollection at: startIndex
	"Write anIntegers worth of elements from aSequenceableCollection starting at startIndex into self."
	"	anInteger	<Integer>	the number of elements to write
	 	aStreamable	<SequenceableCollection>	the source to write in to the destination
	 	startIndex	<Integer>	the index into aSequenceableCollection to start writing from
		^<Integer>	number of elements written
	""
		String new writing write: 3 from: 'Hello' at: 2; close; destination
	"
	self subclassResponsibility! !

!XTWriteStream methodsFor: 'transforming' stamp: 'nice 10/9/2010 20:04'!
writing
	"Return a writeable stream. This is to make the method polymorphic between write streams and terminals."
	"	^	<WriteStream>
	"
	^self! !

!SequenceableCollection methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:13'!
streamingInsert: anInteger into: aWriteStream
	aWriteStream insert: anInteger from: self at: 1! !

!SequenceableCollection methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:13'!
streamingInsertInto: aWriteStream
	aWriteStream insert: self size from: self at: 1.
	^self size! !

!SequenceableCollection methodsFor: '*Xtreams-Core-private' stamp: 'MartinKobetic 12/30/2010 22:33'!
streamingWrite: anInteger into: aWriteStream
	^aWriteStream write: anInteger from: self at: 1! !

!SequenceableCollection methodsFor: '*Xtreams-Core-private' stamp: 'nice 10/9/2010 19:13'!
streamingWriteInto: aWriteStream
	aWriteStream write: self size from: self at: 1.
	^self size! !
XTWriteStream initialize!
